c2345678911234567892123456789312345678941234567895123456789612345678971234567898
      program iam2fits
c
c  Takes co-ordinates and size and extracts IAM data into a binary FITS 
c  table file along with header info from the house-keeping file 
c
c  Compile: 
c
c  % f77 iam2fits.f -L/star/lib `sla_link` `fitsio_link` -o iam2fits -vms (OSF)
c                                                                    -xl 
c                                                                     (Solaris)
c
c  NB: For unswapped data on Suns, change parameter swap to true before
c      compilation.
c
      implicit none
      logical swap,ok,zhflags(2),maxflg
      parameter (swap=.false.)
c      parameter (swap=.true.)
      character line*80,path*80
      integer i,j,size,nstrt,adim
      integer mdim,status,strt,buff(32),length,buffra,imax
      integer inx,iny,num,equinox,osize
      parameter (adim=1152,mdim=131072,inx=4096,iny=inx)
      character iamfile*80,output*80
      real side,dr
      integer ra,dec,points(mdim),idelra,ideldc,i2pi
      double precision pi
      double precision reslt,dra,ddec,ra2000,dec2000,df
c  FITSIO declarations:
      integer unitno,blocksize,bitpix,naxis,naxes(2)
      logical simple,extend

      pi=4.0d0*atan(1.0d0)
      i2pi=nint(pi*2.0d8)

      write (*,*) 'Path-name of measurement:'
      read (*,'(a)') path
      write (*,*) 'Size of square region to extract:'
      write (*,*) 
     :'(arcmin; 0.0 to convert entire IAM file to FITS format)'
      read (*,*) side
      if (side.gt.0.0) then
       write (*,*) 'Centre co-ordinates of region to be extracted:'
       read (*,'(a)') line
       write (*,*) 'Equinox of co-ordinates (0=1900, 1=B1950, 2=J2000):'
1      read (*,*) equinox
       if (equinox.lt.0.or.equinox.gt.2) then
        write (*,*) 'Invalid reply: please answer 0, 1 or 2:'
        goto 1
       endif
      endif
      write (*,*) 'Output FITS filename:'
      read (*,'(a)') output

c  Test data:
c      path='/disk1/nch/UKR444'
c      path='/net/cosaxp1/disk1/nch/UKR444'
c      line='13 35 0.593 -27 30 30.00'
c      equinox=2
c      side=2.86464
c      path='/data1/nch/UKR444'
c      side=2.0
c      output='test1.fits'

      size=nint(side*60.0/0.6714)
      osize=size
      dr=side*60.0

c  Set up the filenames:
      i=length(path)
      iamfile=path
      if (path(i:i).eq.'/') then 
       iamfile(i+1:i+9)='iam.srtrd'
      else
       iamfile(i+1:i+10)='/iam.srtrd'
      endif

      open (unit=1,file=iamfile,status='old',access='direct',
     :form='unformatted',recl=32,readonly)

      num=0

      if (side.gt.0.0) then

      strt=1
      nstrt=strt
      call sla_dafin(line,nstrt,reslt,j)
      if (j.lt.0.and.j.gt.-3) then
       write (*,*) 'SLA_DAFIN error decoding RA:'
       write (*,'(a,a)') ' Offending line: '//line
       stop
      else if (j.eq.-3) then
       write (*,*) 'Warning: range check failed for seconds on line'
       write (*,'(1x,a)') line
      endif
      dra=reslt*15.0d0
      call sla_dafin(line,nstrt,reslt,j)
      if (j.lt.0.and.j.gt.-3) then
       write (*,*) 'SLA_DAFIN error decoding DEC:'
       write (*,'(a,a)') ' Offending line: '//line
       stop
      else if (j.eq.-3) then
       write (*,*) 'Warning: range check failed for arcseconds on line'
       write (*,'(1x,a)') line
      endif
      ddec=reslt
      if (equinox.eq.0) then
       call sla_preces('FK4',1900.0d0,1950.0d0,dra,ddec)
       call sla_fk45z(dra,ddec,1950.0D0,ra2000,dec2000)
       ra=nint(ra2000*1.0d8)
       dec=nint(dec2000*1.0d8)
      else if (equinox.eq.1) then
       call sla_fk45z(dra,ddec,1950.0D0,ra2000,dec2000)
       ra=nint(ra2000*1.0d8)
       dec=nint(dec2000*1.0d8)
      else
       ra=nint(dra*1.0d8)
       dec=nint(ddec*1.0d8)
      endif

c  Set up the 0h flags: if the requested area crosses the boundary at RA=0h,
c  then set the flag to true:
      df=dble(dr)*pi*1.0d8/(180.0d0*3600.0d0)
      idelra=nint(df/cos(dble(dec)/1.0d8))
      if ((ra+idelra).ge.nint(2.0d8*pi)) then
       zhflags(1)=.false.
       zhflags(2)=.true.
      else if ((ra-idelra).lt.0) then
       zhflags(1)=.true.
       zhflags(2)=.false.
      else 
       zhflags(1)=.false.
       zhflags(2)=.false.
      endif

      ideldc=nint(dble(dr)*pi*1.0d8/(180.0d0*3600.0d0))
      i=0
30    i=i+1
      read (1,rec=i,err=40) buff
      if (swap) call swbyte4(buff,32)
      if (buff(29).lt.0) goto 30
c  No need to continue if less than lowest or greater than highest RA:
      idelra=nint(df/cos(dble(buff(2))/1.0d8))
      if ((buff(1)+idelra).lt.ra.and..not.zhflags(2)) goto 30
      if ((buff(1)-idelra).gt.ra.and..not.zhflags(1)) goto 30
c  Note: every time a comparison is made between centre and IAM ra, must
c  check to see if chart straddles 0h boundary and tweak IAM ra so as to 
c  include any objects on the other side of the boundary to the chart centre:
      if (zhflags(1).and.i2pi-buff(1).lt.idelra) then
       buffra=buff(1)-i2pi
      else if (zhflags(2).and.buff(1).lt.idelra) then
       buffra=buff(1)+i2pi
      else
       buffra=buff(1)
      endif
      if (abs(buff(2)-dec).le.ideldc.and.abs(buffra-ra).le.
     :idelra) then
       num=num+1
       if (num.le.mdim) then
        points(num)=i
       else if (num.eq.mdim+1) then
        write (*,*) 
     :'Warning: maximum number of matches exceeded for object ...'
        maxflg=.true.
       endif
      endif
      goto 30
40    if (maxflg) then
       imax=num
       write (*,*) 'Increase dimension MDIM and recompile -'
       write (*,*) 'number should be at least ',imax
       write (*,*) 
     :'No. of images that will be written to FITS file: ',mdim
       num=mdim
      else if (num.eq.0) then
       write (*,*) 'No images were selected in the specified region.'
       stop
      endif

      write (*,*) 'Number of records read from IAM file: ',i-1

      endif

       ok=.true.
     
c  Open the FITS file and write in the header info:
        status=0
        call deletefile(output,status)
        unitno=2
        blocksize=1
        call ftinit(unitno,output,blocksize,status)
c  Create the first HDU indicating a null primary array:
        simple=.true.
        bitpix=16
        naxis=2
        naxes(1)=2
        naxes(2)=2
        extend=.true.
        call ftphpr(unitno,simple,bitpix,1,0,0,1,extend,status)
c  Write the house-keeping information:
        call hkfits(unitno,path,2,status)
        call ftpkye(unitno,'EQUINOX',2000.0,6,
     :'Julian reference frame equinox',status)
c  Write the IAM data:
       call ftcrhd(unitno,status)
       call iamtable(unitno,points,max(1,num),1,status,swap,side)

       call ftclos(unitno,status)
       if (status.gt.0) call printerror(status)

      end
       
      include 'swbyte4.f'
      include 'length.f'
      include 'deletefile.f'
      include 'printerror.f'
      include 'hkfits.f'
      include 'iamtable.f'
