c2345678911234567892123456789312345678941234567895123456789612345678971234567898
      subroutine newfitxy(itype,np,xye,xym,coeffs,jf)
c
c  New version of least-squares zp/scale/rot utilising SLA_FITXY directly for
c  ITYPE = 4 or 6, but also including new ITYPE = 3 for a simple shift/rotate
c  with no scale change, ITYPE = 2 for zero point shift only, and ITYPE = 0
c  to do nothing.
c
      implicit none
      INTEGER ITYPE,NP
      DOUBLE PRECISION XYE(2,NP),XYM(2,NP),COEFFS(6)
      double precision prrms,mrrms,xz,yz,xs,ys,perp,orient,phi
      double precision xtot,ytot,p,xp,yp,pi
      parameter (pi=3.141592653589793)
      double precision a(3,3),b(3),mk,ck1,ck2,ck3,det,theta
      INTEGER i,j,jf,iw(3),flag
      logical sth
c  Floating point traps - uncomment for debugging under Solaris:
c      integer ieeer,ieee_handler
c      external hdl
c trap common floating point exceptions:
c      ieeer = ieee_handler('set','common',hdl)


*  Preset the status
      Jf=0

*  Float the number of samples
      P=DBLE(NP)

      if (itype.lt.0.or.itype.gt.6.or.itype.eq.1.or.itype.eq.5) then

       jf=-1

      else

       if (itype.eq.4.or.itype.eq.6) then

        call sla_fitxy(itype,np,xye,xym,coeffs,jf)

       else

        if (itype.eq.0) then

         coeffs(1)=0.0d0
         coeffs(2)=1.0d0
         coeffs(3)=0.0d0
         coeffs(4)=0.0d0
         coeffs(5)=0.0d0
         coeffs(6)=1.0d0

        else

*     Check enough samples
         IF (NP.GE.2) THEN
 
          if (itype.eq.2) then

c         calculate the zero points for ITYPE = 2:

           xtot=0.0d0
           ytot=0.0d0
           do i=1,np
            xtot=xtot+(xye(1,i)-xym(1,i))
            ytot=ytot+(xye(2,i)-xym(2,i))
           enddo
           xz=xtot/p
           yz=ytot/p

           coeffs(1)=xz
           coeffs(4)=yz
           coeffs(2)=1.0d0
           coeffs(3)=0.0d0
           coeffs(5)=0.0d0
           coeffs(6)=1.0d0

          else

c  First, determine approximate angle of rotation using ITYPE=4:

           call sla_fitxy(4,np,xye,xym,coeffs,j)
           jf=j

           if (jf.eq.0) then

            call sla_dcmpf(coeffs,xz,yz,xs,ys,perp,orient)
            
            if (abs(orient).lt.pi/4.0d0.or.abs(orient).gt.
     :3.0d0*pi/4.0d0) then
             sth=.true.
            else
             sth=.false.
            endif

c  Initialise sums for three coefficient fit:

            do i=1,3
             do j=1,3
              a(i,j)=0.0d0
             enddo
             b(i)=0.0d0
            enddo

            do i=1,np

             if (sth) then

c  Solve for xz,yz and sin(theta):
              ck1=xym(2,i)
              ck2=-xym(1,i)
              ck3=xym(1,i)**2 + xym(2,i)**2
              mk=xym(2,i)*xye(1,i)-xym(1,i)*xye(2,i)

             else

c  Solve for xz,yz and cos(theta):
              ck1=xym(1,i)
              ck2=xym(2,i)
              ck3=ck1**2 + ck2**2
              mk=xym(1,i)*xye(1,i) + xym(2,i)*xye(2,i)

             endif

             a(1,2)=a(1,2)+ck1*ck2
             a(1,3)=a(1,3)+ck1*ck3
             a(2,2)=a(2,2)+ck2*ck2
             a(2,3)=a(2,3)+ck2*ck3
             a(3,3)=a(3,3)+ck3*ck3
             a(1,1)=a(1,1)+ck1*ck1
             b(1)=b(1)+ck1*mk
             b(2)=b(2)+ck2*mk
             b(3)=b(3)+ck3*mk

            enddo

            a(2,1)=a(1,2)
            a(3,1)=a(1,3)
            a(3,2)=a(2,3)

            call sla_dmat(3,a,b,det,flag,iw)

            if (flag.ne.0) then

             jf=-3

            else

             if (sth) then
              theta=asin(b(3))
             else
              theta=acos(b(3))
             endif
             coeffs(1)=b(1)
             coeffs(2)=cos(theta)
             coeffs(3)=sin(theta)
             coeffs(4)=b(2)
             coeffs(5)=-sin(theta)
             coeffs(6)=cos(theta)

c  Try the fit to get the residuals:

             prrms=0.0d0
             do i=1,np
              xp=coeffs(1)+coeffs(2)*xym(1,i)+coeffs(3)*xym(2,i)
              yp=coeffs(4)+coeffs(5)*xym(1,i)+coeffs(6)*xym(2,i)
              prrms=prrms+(xp-xye(1,i))**2+(yp-xye(2,i))**2
             enddo
             prrms=sqrt(prrms/p)

c  Now try for theta=-theta (cos) or theta = +/- pi - theta (sin):

             if (sth) then
              if (theta.lt.0.0d0) then
               phi=-pi-theta
              else
               phi=pi-theta
              endif
             else
              phi=-theta
             endif
             coeffs(1)=b(1)
             coeffs(2)=cos(phi)
             coeffs(3)=sin(phi)
             coeffs(4)=b(2)
             coeffs(5)=-sin(phi)
             coeffs(6)=cos(phi)
             mrrms=0.0d0
             do i=1,np
              xp=coeffs(1)+coeffs(2)*xym(1,i)+coeffs(3)*xym(2,i)
              yp=coeffs(4)+coeffs(5)*xym(1,i)+coeffs(6)*xym(2,i)
              mrrms=mrrms+(xp-xye(1,i))**2+(yp-xye(2,i))**2
             enddo
             mrrms=sqrt(mrrms/p)

             if (prrms.gt.mrrms) then
              theta=phi
             endif

             coeffs(1)=b(1)
             coeffs(2)=cos(theta)
             coeffs(3)=sin(theta)
             coeffs(4)=b(2)
             coeffs(5)=-sin(theta)
             coeffs(6)=cos(theta)

            endif
            
           endif
 
          endif

         else
 
          j=-2

         endif                                                               


        endif

       endif

      endif

      return
      end

c      integer function hdl(sig,sip,uap)
c      integer sig
c      structure /fault/
c       integer address
c      end structure
c      structure /siginfo/
c       integer si_signo
c       integer si_code
c       integer si_errno
c       record /fault/ fault
c      end structure
c      record /siginfo/ sip
c      write (*,'(a,i2,a,z8)') 'ieee exception code ',sip.si_code,
c     :' occurred at hex address ',sip.fault.address
c      end
